// Copyright (C) 2017-2021  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

// Package task provides handy utilities to define & trace tasks.
package task

import (
	"context"
	"fmt"

	"lab.nexedi.com/kirr/neo/go/internal/log"
	taskctx "lab.nexedi.com/kirr/neo/go/internal/xcontext/task"
)

// Running is syntactic sugar to push new task to operational stack, trace it and
// adjust error return with task prefix.
//
// use like this:
//
//      defer task.Running(&ctx, "my task")(&err)
func Running(ctxp *context.Context, name string) func(*error) {
	return running(ctxp, name)
}

// Runningf is Running with formatting support.
func Runningf(ctxp *context.Context, format string, argv ...interface{}) func(*error) {
	return running(ctxp, fmt.Sprintf(format, argv...))
}

func running(ctxp *context.Context, name string) func(*error) {
	ctx := taskctx.Running(*ctxp, name)
	*ctxp = ctx
	traceBegin(3, ctx)

	return func(errp *error) {
		traceEnd(2, ctx, *errp) // XXX recheck 2

		// NOTE not *ctxp here - as context pointed by ctxp could be
		// changed when this deferred function is run
		taskctx.ErrContext(errp, ctx)
	}
}


// XXX merge TraceBegin+TraceEnd into just ->	`defer Trace(ctx)(&err)` ?
//     (can log return lineno more precisely)

// TraceBegin traces beginning of a task.
func TraceBegin(ctx context.Context) {
	traceBegin(2, ctx)
}

// TraceEnd traces end of a task.
func TraceEnd(ctx context.Context, err error) {
	traceEnd(2, ctx, err)
}


func traceBegin(depth int, ctx context.Context) {
	log.Depth(depth).Info(ctx, "[") // TODO log -> trace, don't put ":" before "["
}

func traceEnd(depth int, ctx context.Context, err error) {
	e := ""
	if err != nil {
		e = fmt.Sprintf(" (%s)", err)
	}
	log.Depth(depth).Info(ctx, "]"+e) // TODO log -> trace, don't put ":" before "]"
}
